/*
 * eth_ecat_pdo_fmmu.cpp
 *
 *  Created on: May 16, 2023
 *      Author: algin
 */

#include "free_rtos/ethernet_industry/CoE/eth_ecat_pdo_fmmu.hpp"

#include <algorithm>

namespace free_rtos {

namespace ecat_pdo_fmmu {

void EthEcatPdoFMMU::init() {
    std::vector<ecat_buffer::EcatBufferSlave>& buffer_slaves = ecat_buffer_.get_buffer_slaves();

    pdo_fmmu_slaves_.reserve(buffer_slaves.size());

    for(ecat_buffer::EcatBufferSlave& buffer_slave : buffer_slaves) {
        pdo_fmmu_slaves_.emplace_back(EcatPdoFMMUSlave{buffer_slave});
    }
}

void EthEcatPdoFMMU::wait_op() {
    free_rtos::Semaphore& init_sem = ecat_buffer_.get_ecat().get_init_sem();
    free_rtos::Semaphore& process_sem = ecat_buffer_.get_ecat().get_process_sem();

    process_sem.pend();

    uint32_t logical_full_length_write = ecat_buffer_.get_fmmu_global_properties().logical_full_length_write;
    uint32_t logical_full_length_read = ecat_buffer_.get_fmmu_global_properties().logical_full_length_read;
    std::vector<uint8_t> process_data(std::min(logical_full_length_write, logical_full_length_read));

    read(0, process_data);
/*
    for(uint8_t& byte : process_data) {
        DebugP_log("0x%01x", byte);
    }
    DebugP_log("\r\n");
*/
    write(0, process_data);

    init_sem.post();
    process_sem.pend();

    read(0, process_data);

    for(uint32_t i = 0; i < 250; i++) {
        read_write(0, 0, process_data);

        //read(0, process_data);
/*
        for(uint8_t& byte : process_data) {
            DebugP_log("0x%01x", byte);
        }
        DebugP_log("\r\n");
*/
        //write(0, process_data);
    }

    init_sem.post();
}

void EthEcatPdoFMMU::process_write_queue(uint8_t* process_data, uint32_t len) {
    //DebugP_log("write lock\r\n");
    mutex_write_.lock();

    auto queue = queue_write_; // Копия очереди
    queue_write_.clear();

    //DebugP_log("write unlock\r\n");
    mutex_write_.unlock();
    //DebugP_log("queue write dequeue first\r\n");
    auto next = queue.dequeue();

    while(next != nullptr) {
        //DebugP_log("set value\r\n");
        next->set_value(process_data, len);
        //DebugP_log("queue write dequeue\r\n");
        next = queue.dequeue();
    }
}

void EthEcatPdoFMMU::process_read_queue(uint8_t* process_data, uint32_t len) {
    //DebugP_log("read lock\r\n");
    mutex_read_.lock();

    auto queue = queue_read_; // Копия очереди
    queue_read_.clear();

    //DebugP_log("read unlock\r\n");
    mutex_read_.unlock();
    //DebugP_log("queue read dequeue first\r\n");
    auto next = queue.dequeue();

    while(next != nullptr) {
        //DebugP_log("set value\r\n");
        next->set_value(process_data, len);
        //DebugP_log("queue read dequeue\r\n");
        next = queue.dequeue();
    }
}

void EthEcatPdoFMMU::process() {
    wait_op();

    uint32_t logical_full_length_write = ecat_buffer_.get_fmmu_global_properties().logical_full_length_write;
    uint32_t logical_full_length_read = ecat_buffer_.get_fmmu_global_properties().logical_full_length_read;
    std::vector<uint8_t> process_data(std::min(logical_full_length_write, logical_full_length_read));
    free_rtos::Timer& ecat_timer = ecat_buffer_.get_ecat().get_ecat_timer();

    ecat_timer.Wait();

    read(0, process_data);

    while(1) {
        ecat_timer.Wait();
        //DebugP_log("Tick !\r\n");

        read_write(0, 0, process_data);

        //read(0, process_data);

        process_read_queue(process_data.data(), process_data.size());
        process_write_queue(process_data.data(), process_data.size());

        //write(0, process_data);

        pdo_counter_++;
    }
}

void EthEcatPdoFMMU::process_fake(uint32_t period_microsec = 250) {
    std::vector<uint8_t> process_data(110);
    free_rtos::Timer& ecat_timer = ecat_buffer_.get_ecat().get_ecat_timer();

    Timer::Settings ecat_tmr_sett = {
                 .input_clk_Hz = 25000000, // 25MHz
                 .base_address = 0x2400000u, // memory mapping,
                 .clock_src_mux_addr = 0x430081B0u, // sysconfig
                 .int_num = 152u,                   // sysconfig
                 .int_priority = 4,                 // sysconfig
                 .period_us = period_microsec /// microsec
              };

    ecat_timer.Init(ecat_tmr_sett);
    ecat_timer.Start();

    while(1) {
        ecat_timer.Wait();
        //DebugP_log("Tick !\r\n");

        process_read_queue(process_data.data(), process_data.size());
        process_write_queue(process_data.data(), process_data.size());

        pdo_counter_++;
    }

    DebugP_log("process_fake terminated\r\n");
    vTaskDelete(NULL);
}

}

}
