/*
 * eth_ecat_pdo_fmmu.hpp
 *
 *  Created on: May 16, 2023
 *      Author: algin
 */

#ifndef FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_PDO_FMMU_HPP_
#define FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_PDO_FMMU_HPP_

#include "free_rtos/semaphore/semaphore.hpp"
#include "free_rtos/mutex/mutex.hpp"

#include "free_rtos/ethernet_industry/eth_ecat_buffer.hpp"
#include "free_rtos/ethernet_industry/eth_ecat_custom_promise.hpp"

namespace free_rtos {

namespace ecat_pdo_fmmu {

class EcatPdoFMMUSlave {
public:
    EcatPdoFMMUSlave(ecat_buffer::EcatBufferSlave& buffer_slave)
    : buffer_slave_(buffer_slave) { }

    template<typename DataType>
    datagram::EcatDatagram<command::LWR, DataType> make_datagram_write(DataType& data) {
        using TDatagram = datagram::EcatDatagram<command::LWR, DataType>;

        ecat_buffer::FMMUProperties& fmmu_properties = buffer_slave_.get_fmmu_properties_write();
        address::Logical logical = fmmu_properties.address;

        return TDatagram{ {{logical}}, data };
    }

    template<typename DataType>
    datagram::EcatDatagram<command::LRD, DataType> make_datagram_read(DataType& data) {
        using TDatagram = datagram::EcatDatagram<command::LRD, DataType>;

        ecat_buffer::FMMUProperties& fmmu_properties = buffer_slave_.get_fmmu_properties_read();
        address::Logical logical = fmmu_properties.address;

        return TDatagram{ {{logical}}, data };
    }

private:
    ecat_buffer::EcatBufferSlave& buffer_slave_;
};


class EthEcatPdoFMMU {
public:
    EthEcatPdoFMMU(ecat_buffer::EthEcatBuffer& ecat_buffer): ecat_buffer_{ecat_buffer} { }

    void init();
    void process();
    void process_fake(uint32_t period_microsec);

    uint32_t get_pdo_counter() {
        return pdo_counter_;
    }

    template<typename... DataTypes>
    bool pdo_write(uint32_t timeout_ticks, address::Offset offset, DataTypes&... data) {
        custom_promise::WritePromise<DataTypes...> promise{offset, data...};
        auto future = promise.get_future();

        mutex_write_.lock();

        queue_write_ + promise;

        mutex_write_.unlock();

        future.get(timeout_ticks);

        return future.is_ready();
    }

    template<typename... DataTypes>
    bool pdo_read(uint32_t timeout_ticks, address::Offset offset, DataTypes&... data) {
        custom_promise::ReadPromise<DataTypes...> promise{offset, data...};
        auto future = promise.get_future();

        mutex_read_.lock();

        queue_read_ + promise;

        mutex_read_.unlock();

        future.get(timeout_ticks);

        return future.is_ready();
    }


    void pdo_write_async(custom_promise::IPromise& promise) {
        mutex_write_.lock();

        queue_write_ + promise;

        mutex_write_.unlock();
    }

    void pdo_read_async(custom_promise::IPromise& promise) {
        mutex_read_.lock();

        queue_read_ + promise;

        mutex_read_.unlock();
    }

private:
    ecat_buffer::EthEcatBuffer& ecat_buffer_;

    std::vector<EcatPdoFMMUSlave> pdo_fmmu_slaves_;

    Mutex mutex_write_;
    Mutex mutex_read_;

    queue::Queue<custom_promise::IPromise> queue_write_;
    queue::Queue<custom_promise::IPromise> queue_read_;

    uint32_t pdo_counter_{0};

    void wait_op();

    void process_write_queue(uint8_t* process_data, uint32_t len);
    void process_read_queue(uint8_t* process_data, uint32_t len);

    template<typename... DataTypes>
    void write(address::Offset offset, DataTypes&... data) {
        using TDatagram = datagram::EcatDatagram<command::LWR, DataTypes...>;

        telegram::EcatTelegram& telegram = ecat_buffer_.get_ecat().get_telegram();
        ecat_buffer::FMMUGlobalProperties& fmmu_global_properties = ecat_buffer_.get_fmmu_global_properties();
        address::Logical logical = fmmu_global_properties.logical_start_address + offset;
        TDatagram datagram{ {{logical}}, data... };

        do {
            telegram.transfer(datagram);
        } while(datagram.get_wkc() < 0x0001);
    }

    template<typename... DataTypes>
    void read(address::Offset offset, DataTypes&... data) {
        using TDatagram = datagram::EcatDatagram<command::LRD, DataTypes...>;

        telegram::EcatTelegram& telegram = ecat_buffer_.get_ecat().get_telegram();
        ecat_buffer::FMMUGlobalProperties& fmmu_global_properties = ecat_buffer_.get_fmmu_global_properties();
        address::Logical logical = fmmu_global_properties.logical_start_address + fmmu_global_properties.logical_full_length_write + offset;
        TDatagram datagram{ {{logical}}, data... };

        do {
            telegram.transfer(datagram);
        } while(datagram.get_wkc() < 0x0001);
    }

    template<typename... DataTypes>
    void read_write(address::Offset offset_read, address::Offset offset_write, DataTypes&... data) {
        using TDatagramRead = datagram::EcatDatagram<command::LRD, DataTypes...>;
        using TDatagramWrite = datagram::EcatDatagram<command::LWR, DataTypes...>;

        telegram::EcatTelegram& telegram = ecat_buffer_.get_ecat().get_telegram();
        ecat_buffer::FMMUGlobalProperties& fmmu_global_properties = ecat_buffer_.get_fmmu_global_properties();

        address::Logical logical_read = fmmu_global_properties.logical_start_address + fmmu_global_properties.logical_full_length_write + offset_read;
        TDatagramRead datagram_read{ {{logical_read}}, data... };

        address::Logical logical_write = fmmu_global_properties.logical_start_address + offset_write;
        TDatagramWrite datagram_write{ {{logical_write}}, data... };

        auto queue = datagram_read + datagram_write;

        do {
            telegram.transfer(queue);
        } while(datagram_read.get_all_wkc() < 0x0001);
    }

};

} // namespace ecat_pdo_fmmu

} // namespace free_rtos

#endif /* FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_PDO_FMMU_HPP_ */
